import java.io.*;
import ij.*;
import ij.io.*;
import ij.gui.*;
import ij.plugin.*;
import ij.process.*;
import java.util.*;
import java.io.*;

/** ESRIASCIIGrid_Reader
 *
 * Reader plugin for ESRI ASCII raster grid format.
 *
 */

public class ESRIASCIIGrid_Reader implements PlugIn {
	
	ImagePlus img;
	
	private static class ESRIASCIIGridHeader {
		
        protected int cols = 0;
        protected int rows = 0;
        protected double west = Double.NaN;
        protected double south = Double.NaN;
        protected double cellSize = Double.NaN;
        protected float noDataValue = Float.NaN;
		
        /*
         * returns whether valid values have been found
         */
        protected boolean isValid() {
			
            return (cols > 0 
                    && rows > 0
                    && cellSize > 0
                    && !Double.isNaN(west)
                    && !Double.isNaN(south));
            // noDataValue is optional
        }
		
        /**
         * Reads cols, rows, west, south, cellSize and noDataValue from the header.
         * Throws an exception if this is not a valid header.
         * @param scanner Scanner must be initialized to use dot as decimal separator.
         * @throws IOException
         */
        private void readHeader(Scanner scanner) throws IOException {
			
            cols = rows = 0;
            west = south = cellSize = Double.NaN;
            noDataValue = Float.NaN;
			
            while (scanner.hasNext()) {
				
                if (scanner.hasNextDouble()) {
                    // next line starts with number, must be grid
                    break;
                }
				
                String str = scanner.next().trim().toLowerCase();
                if (str.equals("ncols")) {
                    this.cols = scanner.nextInt();
                } else if (str.equals("nrows")) {
                    this.rows = scanner.nextInt();
                } else if (str.equals("xllcenter") || str.equals("xllcorner")) {
                    this.west = scanner.nextDouble();
                } else if (str.equals("yllcenter") || str.equals("yllcorner")) {
                    this.south = scanner.nextDouble();
                } else if (str.equals("cellsize")) {
                    this.cellSize = scanner.nextDouble();
                } else if (str.startsWith("nodata")) {
                    this.noDataValue = scanner.nextFloat();
                } else {
					
                    // make sure the line starts with a number
                    if (!scanner.hasNextDouble()) {
                        throw new IOException();
                    }
					
                    // done reading the header
                    break;
                }
            }
			
            if (!isValid()) {
                throw new IOException();
            }
        }
    }
	
	public ESRIASCIIGrid_Reader() {
		img = null;
	}
	
	 /**
     * Returns whether a scanner references valid data that can be read.
     * @param scanner
     * @return
     * @throws IOException
     */
    public static boolean canRead(Scanner scanner) {
        try {
            ESRIASCIIGridHeader header = new ESRIASCIIGridHeader();
            header.readHeader(scanner);
            return header.isValid();
        } catch (Exception exc) {
            return false;
        }
    }
    
	public static boolean canRead(String filePath) {
        Scanner scanner = null;
        try {
            scanner = createUSScanner(new FileInputStream(filePath));
            return canRead(scanner);
        } catch (Exception exc) {
            return false;
        } finally {
            if (scanner != null) {
                try {
                    scanner.close();
                } catch (Throwable exc) {
                }
            }
        }
    }
    
	public void run(String arg) {
		
		OpenDialog od = new OpenDialog("Open ESRI grid ...", arg);
		String directory = od.getDirectory();
		String fileName = od.getFileName();
		if (fileName==null) return;
		
		IJ.showStatus("Opening: " + directory + fileName);
        
		read(directory,fileName);
		
		if (img==null) return;
		
		img.show();      
		
	}
	
	
protected void read(String dir, String filename) {
	Scanner scanner = null;
        try {
            BufferedReader br = new BufferedReader(new FileReader(dir+filename));
			File file = new File(dir+filename);
    	    FileInputStream fis = new FileInputStream(file.getAbsolutePath());
			scanner = createUSScanner(new BufferedInputStream(fis));       
			ESRIASCIIGridHeader header = new ESRIASCIIGridHeader();
            header.readHeader(scanner);

            int height = header.rows;
			int width = header.cols;
			img = NewImage.createFloatImage(filename, width, height, 1, NewImage.FILL_BLACK);

            // use legacy StringTokenizer, which is considerably faster than
            // the Scanner class, which uses regular expressions.
            StringTokenizer tokenizer = new StringTokenizer(scanner.nextLine(), " ");

			float[] pixels = (float[]) img.getProcessor().getPixels();
			float minimum = Float.MAX_VALUE;
			float maximum = -Float.MAX_VALUE;
			for (int row=0;row<height;row++) {
				int offset = row*width;
				for (int col = 0; col < width; col++) {
					// a logical row in the grid does not necesseraly correspond
                    // to a line in the file!
                    if (!tokenizer.hasMoreTokens()) {
                        tokenizer = new StringTokenizer(scanner.nextLine(), " ");
                    }
                    final float v = Float.parseFloat(tokenizer.nextToken());
                    if (v == header.noDataValue || Float.isNaN(v)) {
                       pixels[offset+col] = 0; // Float.NaN;
                    } else {
                        pixels[offset+col] = v;
                        if (v < minimum) {
                        	minimum = v;
                        }
                        if (v > maximum) {
                        	maximum = v;
                        }
                    }
                }
			}
			
			((ij.process.FloatProcessor)(img.getProcessor())).setMinAndMax(minimum, maximum); 

			
        } catch (Exception e) {
			IJ.error("Simple ASCII Reader", e.getMessage());
		} finally {
            try {
                // this closes the input stream
                if (scanner != null) {
                	scanner.close();
               	}
            } catch (Exception exc) {
            }
        }
	}
	
	/**
     * Creates a scanner for ASCII text with a period as decimal separator.
     * @param is
     * @return
     * @throws FileNotFoundException
     */
    private static Scanner createUSScanner(InputStream is) throws FileNotFoundException {
        Scanner scanner = new Scanner(is, "US-ASCII");
        scanner.useLocale(Locale.US);
        return scanner;
    }
	
}
